/*
 * Sweep, a sound wave editor.
 *
 * Copyright (C) 2000 Conrad Parker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#include <sys/ioctl.h>
#include <pthread.h>

#include <sweep/sweep_types.h>
#include <sweep/sweep_sample.h>
#include <sweep/sweep_typeconvert.h>

#include "play.h"
#include "driver.h"

#include "sample-display.h"

#define SCRUB_LATENCY 2.0

static sw_handle * handle = NULL;

static int playoffset = 0, realoffset = 0;
static sw_sample * prev_sample = NULL;
static pthread_t player_thread = -1;

/*
 * update_playmarker ()
 *
 * Update the position of the playback marker line for the sample
 * being played.
 *
 * Called within the main sweep interface thread.
 * gtk_idle will keep calling this function as long as this sample is
 * playing, unless otherwise stopped.
 */
static gint
update_playmarker (gpointer data)
{
  sw_sample * s = (sw_sample *)data;

  if (s->play_mode == SWEEP_TRANSPORT_STOP) {

#ifdef DEBUG
    g_print ("update_playmarker: play_mode is STOP\n");
#endif
    s->playmarker_tag = 0;

    /* Set user offset to correct offset */
    if (s->previewing) {
      sample_set_playmarker (s, s->stop_offset, TRUE);
      sample_set_previewing (s, FALSE);
    } else {
      sample_set_playmarker (s, s->play_offset, TRUE);
    }
    
    /* As this may have been stopped by the player thread, refresh the
     * interface */
    sample_set_playmode (s, SWEEP_TRANSPORT_STOP);
    
    return FALSE;
  } else {
    sample_set_playmarker (s, realoffset, FALSE);

    return TRUE;
  }
}

static void
start_playmarker (sw_sample * s)
{
  if (s->playmarker_tag > 0) {
    gtk_timeout_remove (s->playmarker_tag);
  }

  s->playmarker_tag =
    gtk_timeout_add ((guint32)30,
		     (GtkFunction)update_playmarker,
		     (gpointer)s);
}

static struct timeval tv_instant = {0, 0};

#define FORWARDS (1)
#define BACKWARDS (0)

/* returns TRUE if it plays all the way to end */
static gboolean
play_view(sw_view * view, sw_framecount_t start, sw_framecount_t end,
	  sw_framecount_t lbound, sw_framecount_t ubound,
	  gfloat relpitch)
{
  sw_sample * s = view->sample;
  sw_format * format;
  fd_set fds;
  ssize_t n = 0;
  sw_audio_t * d;
  gint sbytes, channels;
  gdouble po = 0.0, p, startf, endf;
  gdouble momentum_delta = 0.0;
  gboolean do_smoothing = FALSE;
  gint i=0, si=0;
  gint direction = FORWARDS;
  sw_framecount_t last_user_offset = -1;
  sw_audio_t pbuf[PBUF_SIZE];

  if (handle == NULL) return FALSE;

  d = (sw_audio_t *)s->sounddata->data;

  sbytes = 2;
 
  format = s->sounddata->format;
  channels = format->channels;

  /* compensate for drivers unable to do required sampling rate */
  relpitch *= (format->rate / handle->driver_rate);
  
  startf = (gdouble)(lbound);
  endf = (gdouble)(ubound-channels); /* linear interp wants to look ahead
				      * one frame */

  if (s->play_reverse) {
    playoffset = end;
  } else {
    playoffset = start;
  }

  po = (gdouble)playoffset;

  memset (pbuf, 0, sizeof (pbuf));

  while ((s->play_reverse ? (po >= startf) : (po < endf)) &&
	 sample_bank_contains (s) &&
	 s->play_mode != SWEEP_TRANSPORT_STOP) {
    FD_ZERO (&fds);
    FD_SET (handle->driver_fd, &fds);
    
    if (select (handle->driver_fd+1, NULL, &fds, NULL, &tv_instant) == 0);

    /*    memset (pbuf, 0, sizeof (pbuf));*/

    for (i = 0; i < PBUF_SIZE; i++) {

      if (s->play_mute || s->user_offset == last_user_offset) {
	pbuf[i] = 0;
	if (channels == 1 && handle->driver_channels == 2) {
	  pbuf[++i] = 0;
	}

      } else {

	switch (channels) {
	case 1:
	  switch (handle->driver_channels) {
	  case 1:
	    /* Mono to mono */
	    si = (int)floor(po);
	    p = po - (gdouble)si;
	    pbuf[i] = view->gain * (d[si] * p + d[si+channels] * (1 - p));
	    break;
	  case 2:
	    /* Mix mono data up to stereo device */
	    si = (int)floor(po);
	    p = po - (gdouble)si;
	    pbuf[i] = view->gain *
	      (d[si] * p + d[si+channels] * (1 - p));
	    pbuf[i+1] = pbuf[i];
	    i++;
	    break;
	  default:
	    break;
	  }
	  break;
	default:
	  switch (handle->driver_channels) {
	  case 1:
	    /* Mix multichannel data down to mono device */
	    si = (int)floor(po);
	    p = po - (gdouble)si;
	    si *= channels;
	    pbuf[i] = view->gain * 0.5 * ((d[si] + d[si+1]) * p +
	       (d[si+channels] + d[si+1+channels]) * (1 - p));
	    break;
	  case 2:
	    /* Stereo/Multichannel to stereo */
	    si = (int)floor(po);
	    p = po - (gdouble)si;
	    si *= channels;
	    pbuf[i] = view->gain * (d[si] * p + d[si+channels] * (1 - p));
	    i++; si++;
	    pbuf[i] = view->gain * (d[si] * p + d[si+channels] * (1 - p));
	    break;
	  default:
	    break;
	  }
      
	  break;
	}

#if 1
	/* If this data came about from scrubbing, it is probably a bit
	 * noisy; run it through a simple smoothing filter:
	 *
	 *   x[n] = ( x[n] + 2 * x[n-1] + 3 * x[n-2] ) / 6
	 */
	if (do_smoothing) {
	  int iL1, iL2;
	  int iR1, iR2;
	  
	  switch (handle->driver_channels) {
	  case 1:
	    iL1 = (i - 1 + PBUF_SIZE) % PBUF_SIZE;
	    iL2 = (i - 2 + PBUF_SIZE) % PBUF_SIZE;
	    pbuf[i] += pbuf[iL1]*2.0 + pbuf[iL2]*3.0;
	    pbuf[i] /= 6.0;
	    break;
	  case 2:
	    iL1 = (i - 2 + PBUF_SIZE) % PBUF_SIZE;
	    iL2 = (i - 4 + PBUF_SIZE) % PBUF_SIZE;
	    iR1 = (i - 3 + PBUF_SIZE) % PBUF_SIZE;
	    iR2 = (i - 5 + PBUF_SIZE) % PBUF_SIZE;
	    pbuf[i] += pbuf[iL1] * 2.0 + pbuf[iL2] * 3.0;
	    pbuf[i] /= 6.0;
	    pbuf[i-1] += pbuf[iR1] * 2.0 + pbuf[iR2] * 3.0;
	    pbuf[i-1] /= 6.0;
	  }
	  do_smoothing = FALSE;
	}
#endif
      }


      if (s->by_user) {
	if (s->play_scrubbing) {
	  gdouble old_po = po;
	  gdouble play_delta;

	  if (s->user_offset < (int)po) {
	    direction = BACKWARDS;
	  } else {
	    direction = FORWARDS;
	  }

#if 0
	  po = (po + (SCRUB_LATENCY-1.0)*(gdouble)s->user_offset) /
	    SCRUB_LATENCY;
#elif 1
	  po = (po*(SCRUB_LATENCY-1.0) + (gdouble)s->user_offset) /
	    SCRUB_LATENCY;
#else
	  po = (gdouble)s->user_offset;
#endif

	  momentum_delta = (po - old_po)/8;
	      
	  play_delta = view->rate * relpitch;
	  play_delta += momentum_delta;

	  if (direction == FORWARDS && !s->play_reverse)
	    po += play_delta;
	  else if (direction == BACKWARDS && s->play_reverse)
	    po -= play_delta;

	  playoffset = (int)po;

	} else {
	  playoffset = s->user_offset;
	  po = (gdouble)playoffset;
	  direction = FORWARDS;
	}

	last_user_offset = s->user_offset;

	s->by_user = FALSE;
	do_smoothing = TRUE;
      } else {
	gdouble play_delta;

	play_delta = view->rate * relpitch;

	if (s->play_reverse) {
	  po -= play_delta;
	} else {
	  po += play_delta;
	}

	po += momentum_delta;

	playoffset = (int)po;

	momentum_delta *= 0.86;
      }

      if (po < startf || po >= endf) break;
    } /* for (i) */

    /* Only write if still playing */
    if (s->play_mode != SWEEP_TRANSPORT_STOP) {

      n = device_write (handle, pbuf, PBUF_SIZE, playoffset);

      g_mutex_lock (s->play_mutex);

      realoffset = device_offset (handle);
      if (realoffset == -1) {
	realoffset = playoffset;
      }

      s->play_offset = realoffset;
      
      if (s->by_user && s->play_scrubbing) {
	playoffset = s->user_offset;
      } else {
	s->user_offset = realoffset;
      }

      g_mutex_unlock (s->play_mutex);
    }

  }

  return (s->play_reverse ? (po <= startf) : (po >= endf));
}

static void
init_playback (sw_sample * s)
{
  g_mutex_lock (s->play_mutex);

  s->play_offset = s->user_offset;
  s->by_user = FALSE;

  if (s->play_offset >= s->sounddata->nr_frames) {
    s->play_offset = 0;
  }
  g_mutex_unlock (s->play_mutex);
}

static void
pva (sw_view * view)
{
  sw_sample * s = view->sample;
  sw_framecount_t nr_frames, start, end;

  device_setup (handle, s);

  init_playback (s);
 
  do {

    nr_frames = s->sounddata->nr_frames;

    if (s->play_reverse) {
      start = 0;
      end = s->play_offset;
    } else {
      start = s->play_offset;
      end = nr_frames;
    }

    if (play_view (view, start, end, 0, nr_frames, 1.0)) {
      g_mutex_lock (s->play_mutex);
      if (s->play_looping) {
	s->play_offset = s->play_reverse ? nr_frames : 0;
      } else {
	s->play_mode = SWEEP_TRANSPORT_STOP;
	s->play_offset = s->stop_offset;
      }
      g_mutex_unlock (s->play_mutex);
    }
    
  } while (s->play_mode == SWEEP_TRANSPORT_PLAY && s->play_looping);

  if (!s->play_looping) {
    device_drain (handle);

    g_mutex_lock (s->play_mutex);
    s->play_mode = SWEEP_TRANSPORT_STOP;
    g_mutex_unlock (s->play_mutex);
  }

  device_reset (handle);
  device_close (handle);
}

void
play_view_all (sw_view * view)
{
  sw_handle * h;

  sw_sample * s = view->sample;

  stop_playback (prev_sample);

  sample_set_playmode (s, SWEEP_TRANSPORT_PLAY);
  sample_set_stop_offset (s);
  sample_set_previewing (s, FALSE);

  prev_sample = s;

  if ((h = device_open(O_WRONLY)) != NULL) {
    handle = h;
#ifdef DEBUG
    if (player_thread != -1) {
      g_print ("Warning: runaway player thread!\n");
    }
#endif
    pthread_create (&player_thread, NULL, (void *) (*pva), view);
    start_playmarker (s);
  } else {
    sample_set_playmode (s, SWEEP_TRANSPORT_STOP);
  }
}

static void
pvpreview_cursor (sw_view * view)
{
  sw_sample * s = view->sample;
  sw_framecount_t delta, preview_start, preview_end;

  device_setup (handle, s);

  init_playback (s);
  
  delta = time_to_frames (s->sounddata->format, 1.0); /* 1 second */ 

  preview_start = s->play_offset - delta;
  preview_start = MAX (preview_start, 0);

  preview_end = s->play_offset + delta;
  preview_end = MIN (preview_end, s->sounddata->nr_frames);

  if (play_view (view, preview_start, preview_end,
		 preview_start, preview_end, 1.0)) {
    g_mutex_lock (s->play_mutex);
    s->play_offset = s->stop_offset;
    g_mutex_unlock (s->play_mutex);
  }

  device_drain (handle);

  g_mutex_lock (s->play_mutex);
  s->play_mode = SWEEP_TRANSPORT_STOP;
  g_mutex_unlock (s->play_mutex);

  device_reset (handle);
  device_close (handle);
}

void
play_preview_cursor (sw_view * view)
{
  sw_handle * h;

  sw_sample * s = view->sample;

  stop_playback (prev_sample);

  sample_set_playmode (s, SWEEP_TRANSPORT_PLAY);
  sample_set_stop_offset (s);
  sample_set_previewing (s, TRUE);

  prev_sample = s;

  if ((h = device_open(O_WRONLY)) != NULL) {
    handle = h;
#ifdef DEBUG
    if (player_thread != -1) {
      g_print ("Warning: runaway player thread!\n");
    }
#endif
    pthread_create (&player_thread, NULL, (void *) (*pvpreview_cursor),
		    view);
    start_playmarker (s);
  } else {
    sample_set_playmode (s, SWEEP_TRANSPORT_STOP);
  }
}

static void
pvpreroll (sw_view * view)
{
  sw_sample * s = view->sample;
  sw_framecount_t delta, preroll_start, preroll_end;

  device_setup (handle, s);

  init_playback (s);

  delta = time_to_frames (s->sounddata->format, 1.0); /* preroll 1 second */
 
  preroll_start = s->play_reverse ? s->play_offset : s->play_offset - delta;
  preroll_start = MAX (preroll_start, 0);

  preroll_end = s->play_reverse ? s->play_offset + delta : s->play_offset;
  preroll_end = MIN (preroll_end, s->sounddata->nr_frames);

  if (play_view (view, preroll_start, preroll_end,
		 preroll_start, preroll_end, 1.0)) {
    g_mutex_lock (s->play_mutex);
    s->play_offset = s->stop_offset;
    g_mutex_unlock (s->play_mutex);
  }

  device_drain (handle);

  g_mutex_lock (s->play_mutex);
  s->play_mode = SWEEP_TRANSPORT_STOP;
  g_mutex_unlock (s->play_mutex);

  device_reset (handle);
  device_close (handle);
}

void
play_preroll (sw_view * view)
{
  sw_handle * h;

  sw_sample * s = view->sample;

  stop_playback (prev_sample);

  sample_set_playmode (s, SWEEP_TRANSPORT_PLAY);
  sample_set_stop_offset (s);
  sample_set_previewing (s, TRUE);

  prev_sample = s;

  if ((h = device_open(O_WRONLY)) != NULL) {
    handle = h;
#ifdef DEBUG
    if (player_thread != -1) {
      g_print ("Warning: runaway player thread!\n");
    }
#endif
    pthread_create (&player_thread, NULL, (void *) (*pvpreroll), view);
    start_playmarker (s);
  } else {
    sample_set_playmode (s, SWEEP_TRANSPORT_STOP);
  }
}

static void
pvpreview_cut (sw_view * view)
{
  sw_sample * s = view->sample;
  GList * gl, * gl_prev, * gl_next;
  sw_sel * sel;
  sw_framecount_t delta, preview_start, preview_end;
  sw_framecount_t start = 0, end = 0;
  gboolean was_reverse;
  gboolean skip = FALSE; /* voodoo */

  device_setup (handle, s);

  init_playback (s);

  delta = time_to_frames (s->sounddata->format, 1.0);

  g_mutex_lock (s->sounddata->sels_mutex);

  sel = (sw_sel *)s->sounddata->sels->data;
  preview_start = sel->sel_start;

  for (gl = s->sounddata->sels; gl->next; gl = gl->next);
  
  sel = (sw_sel *)gl->data;
  preview_end = sel->sel_end;
  g_mutex_unlock (s->sounddata->sels_mutex);

  gl = s->play_reverse ? gl : s->sounddata->sels;
  
  preview_start -= delta;
  preview_start = MAX (preview_start, 0);

  preview_end += delta;
  preview_end = MIN (preview_end, s->sounddata->nr_frames);

  g_mutex_lock (s->play_mutex);
  s->play_offset = s->play_reverse ? preview_end : preview_start;
  g_mutex_unlock (s->play_mutex);

  while (gl) {
    
    if (s->play_mode != SWEEP_TRANSPORT_PLAY) break;
    
    /* Hold sels_mutex for as little time as possible */
    g_mutex_lock (s->sounddata->sels_mutex);
    
    /* if gl is no longer in sels, break out */
    if (g_list_position (s->sounddata->sels, gl) == -1) {
      g_mutex_unlock (s->sounddata->sels_mutex);
      break;
    }
    
    sel = (sw_sel *)gl->data;
    start = sel->sel_start;
    end = sel->sel_end;

    gl_prev = gl->prev;
    gl_next = gl->next;

    g_mutex_unlock (s->sounddata->sels_mutex);

    was_reverse = s->play_reverse;

    if (!skip) {
      if (s->play_reverse) {
	play_view (view, end, s->play_offset, end, s->play_offset, 1.0);
      } else {
	play_view(view, s->play_offset, start, s->play_offset, start, 1.0);
      }

      skip = (was_reverse != s->play_reverse);
    } else {
      skip = FALSE;
    }

    g_mutex_lock (s->play_mutex);

    if (s->play_reverse) {
      s->play_offset = start;
      gl = gl_prev;
    } else {
      s->play_offset = end;
      gl = gl_next;
    }

    g_mutex_unlock (s->play_mutex);

  }

  if (s->play_mode == SWEEP_TRANSPORT_PLAY) {
    if (s->play_reverse) {
      play_view (view, preview_start, start, preview_start, start, 1.0);
    } else {
      play_view (view, end, preview_end, end, preview_end, 1.0);
    }
  }

  g_mutex_lock (s->play_mutex);
  s->play_offset = s->stop_offset;
  g_mutex_unlock (s->play_mutex);
  
  if (s->play_mode == SWEEP_TRANSPORT_PLAY) {
    device_drain (handle);

    g_mutex_lock (s->play_mutex);
    s->play_mode = SWEEP_TRANSPORT_STOP;
    g_mutex_unlock (s->play_mutex);
  }

  device_reset (handle);
  device_close (handle);
}

void
play_preview_cut (sw_view * view)
{
  sw_handle * h;

  sw_sample * s = view->sample;

  if (s->sounddata->sels == NULL) {
    sample_set_playmode (s, SWEEP_TRANSPORT_STOP);
    return;
  }

  stop_playback (prev_sample);

  sample_set_playmode (s, SWEEP_TRANSPORT_PLAY);
  sample_set_stop_offset (s);
  sample_set_previewing (s, TRUE);

  prev_sample = s;

  if ((h = device_open(O_WRONLY)) != NULL) {
    handle = h;
#ifdef DEBUG
    if (player_thread != -1) {
      g_print ("Warning: runaway player thread\n");
    }
#endif
    pthread_create (&player_thread, NULL, (void *) (*pvpreview_cut), view);
    start_playmarker (s);
  } else {
    sample_set_playmode (s, SWEEP_TRANSPORT_STOP);
  }
}

static void
pvs (sw_view * view)
{
  sw_sample * s = view->sample;
  GList * gl, * gl_prev, * gl_next;
  sw_sel * sel;
  sw_framecount_t sels_start, sels_end;
  sw_framecount_t start, end;
  gboolean restart_condition = FALSE;

  device_setup (handle, s);

  init_playback (s);

  g_mutex_lock (s->sounddata->sels_mutex);

  gl = s->sounddata->sels;
  sel = (sw_sel *)gl->data;
  sels_start = sel->sel_start;

  gl = g_list_last (s->sounddata->sels);
  
  sel = (sw_sel *)gl->data;
  sels_end = sel->sel_end;
  g_mutex_unlock (s->sounddata->sels_mutex);

  restart_condition = (s->play_reverse ? s->play_offset <= sels_start :
		       s->play_offset >= sels_end);

  do {

    if (restart_condition) {
      g_mutex_lock (s->play_mutex);
      s->play_offset = s->play_reverse ? sels_end : sels_start;
      g_mutex_unlock (s->play_mutex);
    }

    g_mutex_lock (s->sounddata->sels_mutex);

    gl = s->sounddata->sels;
    sel = (sw_sel *)gl->data;
    sels_start = sel->sel_start;
    
    gl = g_list_last (s->sounddata->sels);
    
    sel = (sw_sel *)gl->data;
    sels_end = sel->sel_end;
 
    gl = s->play_reverse ? gl : s->sounddata->sels;

    g_mutex_unlock (s->sounddata->sels_mutex);
    
    while (gl) {
      
      if (s->play_mode != SWEEP_TRANSPORT_PLAY_SEL) break;

      /* Hold sels_mutex for as little time as possible */
      g_mutex_lock (s->sounddata->sels_mutex);
      
      /* if gl is no longer in sels, break out */
      if (g_list_position (s->sounddata->sels, gl) == -1) {
	g_mutex_unlock (s->sounddata->sels_mutex);
	break;
      }
      
      sel = (sw_sel *)gl->data;
      start = sel->sel_start;
      end = sel->sel_end;

      gl_prev = gl->prev;
      gl_next = gl->next;
      
      g_mutex_unlock (s->sounddata->sels_mutex);

      if (s->play_reverse) {
	if (start < s->play_offset) {
	  if (end > s->play_offset) {
	    end = s->play_offset;
	  } else {
	    g_mutex_lock (s->play_mutex);
	    s->play_offset = end;
	    g_mutex_unlock (s->play_mutex);
	  }
	  restart_condition = play_view (view, start, end, start, end, 1.0);
	}
      } else {
	if (end > s->play_offset) {
	  if (start < s->play_offset) {
	    start = s->play_offset;
	  } else {
	    g_mutex_lock (s->play_mutex);
	    s->play_offset = start;
	    g_mutex_unlock (s->play_mutex);
	  }
	  restart_condition = play_view (view, start, end, start, end, 1.0);
	}
      }
      
      gl = s->play_reverse ? gl_prev : gl_next;
    }

    if (restart_condition && !s->play_looping) {
      g_mutex_lock (s->play_mutex);
      s->play_mode = SWEEP_TRANSPORT_STOP;
      s->play_offset = s->play_reverse ? sels_start : sels_end;
      g_mutex_unlock (s->play_mutex);
    }

  } while (s->play_mode == SWEEP_TRANSPORT_PLAY_SEL &&
	   s->sounddata->sels != NULL &&
	   (s->play_looping || s->play_scrubbing));

  if (s->play_mode == SWEEP_TRANSPORT_PLAY_SEL) {
    device_drain (handle);

    g_mutex_lock (s->play_mutex);
    s->play_mode = SWEEP_TRANSPORT_STOP;
    g_mutex_unlock (s->play_mutex);
  }

  if (playoffset >= s->sounddata->nr_frames)
    sample_set_playmarker (s, 0, FALSE);

  device_reset (handle);
  device_close (handle);
}

void
play_view_sel (sw_view * view)
{
  sw_handle * h;

  sw_sample * s = view->sample;

  if (s->sounddata->sels == NULL) {
    sample_set_playmode (s, SWEEP_TRANSPORT_STOP);
    return;
  }

  stop_playback (prev_sample);

  sample_set_playmode (s, SWEEP_TRANSPORT_PLAY_SEL);
  sample_set_stop_offset (s);
  sample_set_previewing (s, FALSE);

  prev_sample = s;

  if ((h = device_open(O_WRONLY)) != NULL) {
    handle = h;
#ifdef DEBUG
    if (player_thread != -1) {
      g_print ("Warning: runaway player thread\n");
    }
#endif
    pthread_create (&player_thread, NULL, (void *) (*pvs), view);
    start_playmarker (s);
  } else {
    sample_set_playmode (s, SWEEP_TRANSPORT_STOP);
  }
}

typedef struct _pvap_data pvap_data;
struct _pvap_data {
  sw_view * view;
  gfloat pitch;
};

static void
pvap (pvap_data * p)
{
  sw_sample * s = p->view->sample;
  sw_framecount_t nr_frames, start, end;

  device_setup (handle, s);

  init_playback (s);

  nr_frames = s->sounddata->nr_frames;

  if (p->view->current_tool != TOOL_SCRUB ||
      (s->play_reverse ? (s->play_offset <= 0) :
       (s->play_offset >= nr_frames))) {
    g_mutex_lock (s->play_mutex);
    s->play_offset = s->play_reverse ? nr_frames : 0;
    g_mutex_unlock (s->play_mutex);
  }

  do {

    nr_frames = s->sounddata->nr_frames;

    if (s->play_reverse) {
      start = 0;
      end = s->play_offset;
    } else {
      start = s->play_offset;
      end = nr_frames;
    }

    if (play_view (p->view, start, end, 0, nr_frames, p->pitch)) {
      g_mutex_lock (s->play_mutex);
      if (s->play_looping) {
	s->play_offset = s->play_reverse ? nr_frames : 0;
      } else {
	s->play_mode = SWEEP_TRANSPORT_STOP;
	s->play_offset = s->stop_offset;
      }
      g_mutex_unlock (s->play_mutex);
    }
    
  } while (s->play_mode == SWEEP_TRANSPORT_PLAY && s->play_looping);

  if (!s->play_looping) {
    device_drain (handle);

    g_mutex_lock (s->play_mutex);
    s->play_mode = SWEEP_TRANSPORT_STOP;
    g_mutex_unlock (s->play_mutex);
  }

  device_reset (handle);
  device_close (handle);

  g_free (p);
}

void
play_view_all_pitch (sw_view * view, gfloat pitch)
{
  sw_handle * h;
  sw_sample * s = view->sample;
  pvap_data * p;
  sw_framecount_t mouse_offset;

  p = g_malloc (sizeof (*p));
  p->view = view;
  p->pitch = pitch;

  stop_playback (s);

  sample_set_playmode (s, SWEEP_TRANSPORT_PLAY);
  sample_set_stop_offset (s);
  sample_set_previewing (s, FALSE);

  if (view->current_tool == TOOL_SCRUB) {
    mouse_offset =
      sample_display_get_mouse_offset (SAMPLE_DISPLAY(view->display));
    sample_set_playmarker (s, mouse_offset, TRUE);
    sample_set_scrubbing (s, TRUE);
  }

  prev_sample = s;

  if ((h = device_open(O_WRONLY)) != NULL) {
    handle = h;
#ifdef DEBUG
    if (player_thread != -1) {
      g_print ("Warning: runaway player thread\n");
    }
#endif
    pthread_create (&player_thread, NULL, (void *) (*pvap), p);
    start_playmarker (s);
  } else {
    sample_set_playmode (s, SWEEP_TRANSPORT_STOP);
  }
}

void
stop_all_playback (void)
{
  if (prev_sample != NULL)
    sample_set_playmode (prev_sample, SWEEP_TRANSPORT_STOP);

  device_flush (handle);

  pthread_join (player_thread, NULL);
  player_thread = -1;
}

void
pause_playback (sw_sample * s)
{
  if (s == NULL) return;

  if (s->play_mode != SWEEP_TRANSPORT_STOP) {
    sample_set_playmode (s, SWEEP_TRANSPORT_STOP);

    device_flush (handle);
    
    pthread_join (player_thread, NULL);
    player_thread = -1;
  }

  sample_set_stop_offset (s);
}

void
stop_playback (sw_sample * s)
{
  if (s == NULL) return;

  if (s->play_mode != SWEEP_TRANSPORT_STOP) {
    sample_set_playmode (s, SWEEP_TRANSPORT_STOP);

    device_flush (handle);
    
    pthread_join (player_thread, NULL);
    player_thread = -1;

    sample_set_playmarker (s, s->stop_offset, TRUE);
  }
}
